unit FileCollector;

(*

This unit contains classes which collect files.
You can select files via direct input, or via drag'n'drop or command line
parameters. These classes help you making file input easier since it
automatically expands the filenames as well as resoluting environment
variables and wildcards.

Example for an valid input:
.\..\..\*.mp?
%windir%\*.ini

*)

// Future:
// Drag'n'Drop Klasse
// OpenDialog Klasse

interface

uses
  Windows, SysUtils, Classes;

type
  TExpandMode = (emNone, emUNC, emNormal, emCaseSensitive);

  TUniqueStringList = class(TStringList)
  public
    function Add(const S: string): Integer; override;
    function AddObject(const S: String; AObject: TObject): Integer; override;
    procedure Append(const S: String);
    procedure AddStrings(Strings: TStrings); override;
  end;

  TFileNameCollector = class(TObject)
  private
    FExpandMode: TExpandMode;
    function ExpandEnvStr(const szInput: string): string;
  protected
    DirList: TUniqueStringList;
    FileList: TUniqueStringList;
    function GetCollectedFiles(i: integer): string; virtual;
    function GetCollectedFilesCount: integer; virtual;
    function GetCollectedFolders(i: integer): string; virtual;
    function GetCollectedFoldersCount: integer; virtual;
  public
    property CollectedFiles[i: integer]: string read GetCollectedFiles;
    property CollectedFilesCount: integer read GetCollectedFilesCount;
    property CollectedFolders[i: integer]: string read GetCollectedFolders;
    property CollectedFoldersCount: integer read GetCollectedFoldersCount;
    constructor Create(AExpandMode: TExpandMode);
    destructor Destroy; override;
    procedure AddFileOrFolder(const AFile: string); virtual;
    procedure Clear; virtual;
  end;

  TParamFilesCollector = class(TFileNameCollector)
  public
    procedure AddFromParamString(StartFromIndex: integer); overload; virtual;
    procedure AddFromParamString(StartFromIndex, LeaveElementsFromEnd: integer); overload; virtual;
  end;

implementation

{ TUniqueStringList }

function TUniqueStringList.Add(const S: string): Integer;
begin
  result := IndexOf(S);
  if result = -1 then
  begin
    result := inherited Add(S);
  end;
end;

function TUniqueStringList.AddObject(const S: String; AObject: TObject): Integer;
begin
  result := IndexOf(S);
  if result = -1 then
  begin
    result := inherited AddObject(S, AObject);
  end;
end;

procedure TUniqueStringList.Append(const S: String);
begin
  if IndexOf(S) = -1 then
  begin
    inherited Append(S);
  end;
end;

procedure TUniqueStringList.AddStrings(Strings: TStrings);
var
  i: integer;
begin
  for i := 0 to Strings.Count - 1 do
  begin
    if Strings.Objects[i] = nil then
      Add(Strings.Strings[i])
    else
      AddObject(Strings.Strings[i], Strings.Objects[i]);
  end;
end;

{ TFileNameCollector }

function TFileNameCollector.ExpandEnvStr(const szInput: string): string;
const
  MAXSIZE = 32768; // laut PSDK sind 32k das Maximum
begin
  // Ref: http://www.viathinksoft.de/?page=codelib&showid=75
  SetLength(Result,MAXSIZE);
  SetLength(Result,ExpandEnvironmentStrings(pchar(szInput),
    @Result[1],length(Result))-1); //-1 um abschlieendes #0 zu verwerfen
end;

function TFileNameCollector.GetCollectedFiles(i: integer): string;
begin
  result := FileList.Strings[i];
end;

function TFileNameCollector.GetCollectedFilesCount: integer;
begin
  result := FileList.Count;
end;

function TFileNameCollector.GetCollectedFolders(i: integer): string;
begin
  result := DirList.Strings[i];
end;

function TFileNameCollector.GetCollectedFoldersCount: integer;
begin
  result := DirList.Count;
end;

constructor TFileNameCollector.Create(AExpandMode: TExpandMode);
begin
  inherited Create();

  FExpandMode := AExpandMode;

  DirList := TUniqueStringList.Create;
  FileList := TUniqueStringList.Create;
end;

destructor TFileNameCollector.Destroy;
begin
  FileList.Free;
  DirList.Free;

  inherited;
end;

procedure TFileNameCollector.AddFileOrFolder(const AFile: string);
var
  SR: TSearchRec;
  tmp, f: string;
  casematch: TFilenameCaseMatch;
begin
  // Step 1: Expand environment variables

  tmp := ExpandEnvStr(AFile);

  // Step 2: Replace wild cards and divide directories and files.

  if FindFirst(tmp, faAnyFile, SR) = 0 then
  begin
    repeat
      f := SR.Name;

      // Step 3: Expand the file name

      if FExpandMode = emUNC then
        f := ExpandUNCFileName(f)
      else if FExpandMode = emNormal then
        f := ExpandFileName(f)
      else if FExpandMode = emCaseSensitive then
      begin
        f := ExpandFileNameCase(f, casematch);
        // ToDo: Check casematch...
      end;

      // Step 4: Decide if it's a file or a folder

      if DirectoryExists(f) then
        DirList.Add(f)
      else if FileExists(f) then
        FileList.Add(f);
    until FindNext(SR) <> 0;
  end;
  FindClose(SR);
end;

procedure TFileNameCollector.Clear;
begin
  FileList.Clear;
  DirList.Clear;
end;

{ TParamFilesCollector }

procedure TParamFilesCollector.AddFromParamString(StartFromIndex, LeaveElementsFromEnd: integer);
var
  i: integer;
begin
  for i := StartFromIndex to ParamCount {- 1} - LeaveElementsFromEnd do
  begin
    AddFileOrFolder(ParamStr(i));
  end;
end;

procedure TParamFilesCollector.AddFromParamString(StartFromIndex: integer);
begin
  AddFromParamString(StartFromIndex, 0);
end;

end.
