unit vtsSoundUtils;

{.$DEFINE USE_VCL}
{$DEFINE USE_CALLBACK}
{$DEFINE FIXMIDIPOS}

interface

function StartRecording(): boolean; overload;
function StartRecording(BitsPerSample, Channels, SamplesPerSec: integer): boolean; overload;

function StopRecording(): boolean; overload;
function StopRecording(filename: string): boolean; overload;

function playMIDIAndWait(filename: string): boolean;

function getLastMCIError(): string;

{$IFDEF USE_CALLBACK}
type
  TSleepCallback = Function (): boolean;

var
  MIDIPlayingCallback: TSleepCallback;
{$ENDIF}

implementation

uses
  SysUtils,
  {$IFDEF USE_VCL}Forms,{$ENDIF}
  MMSystem,
  Windows;

{$INCLUDE Config.inc}

const
  MCI_REC_ALIAS = 'vtsRecorder';
  MCI_MID_ALIAS = 'vtsMidiPlayer';

var
  lastMCIError: string;
  lastFailedString: string;

function __handle: HWND;
begin
  {$IFDEF USE_VCL}
  result := Application.Handle;
  {$ELSE}
  result := 0;
  {$ENDIF}
end;

{$IFNDEF USE_CALLBACK}
function __Waiting(): boolean;
begin
  result := true;

  {$IFDEF USE_VCL}
  Application.ProcessMessages;
  if Application.Terminated then result := false; // break;
  {$ENDIF}
end;
{$ENDIF}

procedure RaiseError(MCIErr, FailString: string);
begin
  LastMCIError := MCIErr;
  lastFailedString := FailString;
end;

function __mciSendString(lpstrCommand, lpstrReturnString: PChar;
  uReturnLength: UINT; hWndCallback: HWND): boolean;
const
  BUFLEN = 255;
var
  res: MCIERROR;
  s: array[0..BUFLEN] of Char;
begin
  res := mciSendString(lpstrCommand, lpstrReturnString, uReturnLength, hWndCallback);
  if res <> 0 then
  begin
    mciGetErrorString(res, @s, BUFLEN);
    RaiseError(s, lpstrCommand);
    result := false;
  end
  else
  begin
    // RaiseError('', '');
    result := true;
  end;
end;

// Ref: http://www.delphipraxis.net/topic8681,0.html

function StartRecording(): boolean;
begin
  result := StartRecording(BPS, CH, SPS);
end;

function StartRecording(BitsPerSample, Channels, SamplesPerSec: integer): boolean;
begin
  result := false;
  if not __mciSendString('OPEN NEW TYPE WAVEAUDIO ALIAS '+MCI_REC_ALIAS,nil,0,__handle) then exit;
  if not __mciSendString(PChar('SET '+MCI_REC_ALIAS+' ' +
                'TIME FORMAT MS ' +
                'BITSPERSAMPLE '+IntToStr(BitsPerSample)+' ' +
                'CHANNELS '+IntToStr(Channels)+' ' +
                'SAMPLESPERSEC '+IntToStr(SamplesPerSec)+' ' +
                'BYTESPERSEC '+IntToStr((BitsPerSample div 8)*Channels*SamplesPerSec)+' ' +
                'ALIGNMENT '+IntToStr(Channels*(BitsPerSample div 8))),
                nil,0,__handle) then exit;
  if not __mciSendString('RECORD '+MCI_REC_ALIAS,nil,0,__handle) then exit;
  result := true;
end;

function StopRecording(): boolean;
begin
  result := StopRecording('');
end;

function StopRecording(filename: string): boolean;
begin
  result := false;
  if not __mciSendString('STOP '+MCI_REC_ALIAS,nil,0,__handle) then exit;
  if filename <> '' then
  begin
    if not __mciSendString(PChar('SAVE '+MCI_REC_ALIAS+' "'+filename+'"'),nil,0,__handle) then exit;
  end;
  if not __mciSendString('CLOSE '+MCI_REC_ALIAS,nil,0,__handle) then exit;
  result := true;
end;

// Ref: http://www.delphi-library.de/topic_Midi+aus+Ressource+abspielen_70447,0.html

function playMIDIAndWait(filename: string): boolean;
const
  BUFLEN: Cardinal = 255; // MAX_PATH?
var
  BufMIDILen, BufMIDIPos: String;
  MIDILen, MIDIPos, err: integer;
  error: boolean;
resourcestring
  LNG_INTEGER_ERROR = 'Fatal error! Could not get value "%s". The string "%s" is not a valid integer. Maybe the MIDI file is incompatible to your Windows API (MCI).';
  LNG_GENERAL_EXCEPTION = 'Fatal error! Unexpected exception "%s" raised by API. Maybe the MIDI file is not compatbile to your Windows API (MCI).';
begin
  try
    error := false;

    MIDIPos := 0;
    MIDILen := 0;

    result := false;
    if not __mciSendString(PChar('OPEN "SEQUENCER!' + filename + '" ALIAS '+MCI_MID_ALIAS), nil, 0, __handle) then exit;
    if not __mciSendString('PLAY '+MCI_MID_ALIAS, nil, 0, __handle) then exit;

    SetLength(BufMIDILen, BUFLEN);
    // Die nchste Zeile erzeugt EDivByZero at MciSeq.dll bei Windows 95 und SCI Viewer MIDIS (KQ7)
    if not __mciSendString('STATUS '+MCI_MID_ALIAS+' LENGTH', PChar(BufMIDILen), BUFLEN, __handle) then exit;
    repeat
      SetLength(BufMIDIPos, BUFLEN);
      if not __mciSendString('STATUS '+MCI_MID_ALIAS+' POSITION', PChar(BufMIDIPos), BUFLEN, __handle) then exit;
      {$IFDEF USE_CALLBACK}
      if (Assigned(MIDIPlayingCallback)) and not MIDIPlayingCallback() then break;
      {$ELSE}
      if not __Waiting() then break;
      {$ENDIF}

      // Aus irgendeinem Grund habe ich bei einigen MIDI Dateien vom SCI-Viewer
      // hier eine Angabe in Minuten und Stunden anstelle in Sekunden!
      // Windows 2000 strzt im Treiber mit DivByZero ab...
      {$IFDEF FIXMIDIPOS}
      // Dirty Trick:
      bufmidipos := StringReplace(bufmidipos, ':', '', [rfReplaceAll]);
      bufmidilen := StringReplace(bufmidilen, ':', '', [rfReplaceAll]);
      {$ENDIF}

      Val(BufMIDIPos, MIDIPos, err);
      if err <> 0 then
      begin
        RaiseError(Format(LNG_INTEGER_ERROR, ['MIDIPos', BufMIDIPos]), '');
        error := true;
        break;
      end;
      
      Val(BufMIDILen, MIDILen, err);
      if err <> 0 then
      begin
        RaiseError(Format(LNG_INTEGER_ERROR, ['MIDILen', BufMIDILen]), '');
        error := true;
        break;
      end;
    until MIDIPos >= MIDILen;
    if not __mciSendString('STOP '+MCI_MID_ALIAS, nil, 0, __handle) then exit;
    if not __mciSendString('CLOSE '+MCI_MID_ALIAS, nil, 0, __handle) then exit;
    result := not error;
  except
    result := false;
    RaiseError(Format(LNG_GENERAL_EXCEPTION, [ExceptionClass.ClassName]), ''); // ClassName always 'Exception'? (ToDo)
  end;
end;

function getLastMCIError(): string;
resourcestring
  LNG_ERROR = '"%s" at "%s"';
  LNG_UNKNOWN = '(Unknown)';
begin
  if lastFailedString = '' then lastFailedString := LNG_UNKNOWN;
  result := Format(LNG_ERROR, [LastMCIError, lastFailedString]);
end;

end.
